@@ -1,8 +1,7 @@
 # -*- coding: utf-8; mode: tcl; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- vim:fenc=utf-8:ft=tcl:et:sw=4:ts=4:sts=4
-# $Id$
 #
-# Copyright (c) 2002 - 2003 Apple Inc.
-# Copyright (c) 2004 - 2013 The MacPorts Project
+# Copyright (c) 2002-2003 Apple Inc.
+# Copyright (c) 2004-2014, 2016-2018 The MacPorts Project
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -84,20 +83,21 @@ proc portfetch::assemble_url {site distfile} {
 # For a given mirror site type, e.g. "gnu" or "x11", check to see if there's a
 # pre-registered set of sites, and if so, return them.
 proc portfetch::mirror_sites {mirrors tag subdir mirrorfile} {
-    global UI_PREFIX name dist_subdir global_mirror_site
+    global name dist_subdir global_mirror_site
 
     if {[file exists $mirrorfile]} {
         source $mirrorfile
     }
 
     if {![info exists portfetch::mirror_sites::sites($mirrors)]} {
-        if {$mirrors != $global_mirror_site} {
+        if {$mirrors ne $global_mirror_site} {
             ui_warn "[format [msgcat::mc "No mirror sites on file for class %s"] $mirrors]"
         }
         return {}
     }
 
     set ret [list]
+    set name_re {\$(?:name\y|\{name\})}
     foreach element $portfetch::mirror_sites::sites($mirrors) {
 
         # here we have the chance to take a look at tags, that possibly
@@ -114,7 +114,6 @@ proc portfetch::mirror_sites {mirrors tag subdir mirrorfile} {
             set mirror_tag ""
         }
 
-        set name_re {\$(?:name\y|\{name\})}
         # if the URL has $name embedded, kill any mirror_tag that may have been added
         # since a mirror_tag and $name are incompatible
         if {[regexp $name_re $element]} {
@@ -134,13 +133,13 @@ proc portfetch::mirror_sites {mirrors tag subdir mirrorfile} {
             set thesubdir ""
         }
 
-        if {"$tag" != ""} {
-            eval append element "${thesubdir}:${tag}"
+        if {$tag ne ""} {
+            append element "${thesubdir}:${tag}"
         } else {
-            eval append element "${thesubdir}"
+            append element "${thesubdir}"
         }
 
-        eval lappend ret $element
+        lappend ret $element
     }
 
     return $ret
@@ -154,6 +153,8 @@ proc portfetch::mirror_sites {mirrors tag subdir mirrorfile} {
 proc portfetch::checksites {sitelists mirrorfile} {
     global env
     variable urlmap
+    set url_re {([a-zA-Z]+://.+)}
+    set tagged_url_re {([a-zA-Z]+://.+/?):([0-9A-Za-z_-]+)$}
 
     foreach {listname extras} $sitelists {
         upvar #0 $listname $listname
@@ -161,17 +162,22 @@ proc portfetch::checksites {sitelists mirrorfile} {
             continue
         }
         global ${listname}.mirror_subdir
-        # add the specified global and user-defined mirrors
-        set sglobal [lindex $extras 0]; set senv [lindex $extras 1]
         set full_list [set $listname]
-        append full_list " $sglobal"
-        if {[info exists env($senv)]} {
-            set full_list [concat $env($senv) $full_list]
+        # add the specified global and user-defined mirrors
+        set sglobal ""
+        set senv ""
+        if {[llength $extras] >= 2} {
+            set sglobal [lindex $extras 0]
+            set senv [lindex $extras 1]
+            append full_list " $sglobal"
+            if {[info exists env($senv)]} {
+                set full_list [concat $env($senv) $full_list]
+            }
         }
 
         set site_list [list]
         foreach site $full_list {
-            if {[regexp {([a-zA-Z]+://.+)} $site match site]} {
+            if {[regexp $url_re $site match site]} {
                 set site_list [concat $site_list $site]
             } else {
                 set splitlist [split $site :]
@@ -190,19 +196,19 @@ proc portfetch::checksites {sitelists mirrorfile} {
 
         # add in the global and user-defined mirrors for each tag
         foreach site $site_list {
-            if {[regexp {([a-zA-Z]+://.+/?):([0-9A-Za-z_-]+)$} $site match site tag] && ![info exists extras_added($tag)]} {
+            if {[regexp $tagged_url_re $site match site tag] && ![info exists extras_added($tag)]} {
                 if {$sglobal ne ""} {
                     set site_list [concat $site_list [mirror_sites $sglobal $tag "" $mirrorfile]]
                 }
                 if {[info exists env($senv)]} {
-                    set site_list [concat [list $env($senv)] $site_list]
+                    set site_list [concat $env($senv) $site_list]
                 }
                 set extras_added($tag) yes
             }
         }
 
         foreach site $site_list {
-        if {[regexp {([a-zA-Z]+://.+/?):([0-9A-Za-z_-]+)$} $site match site tag]} {
+        if {[regexp $tagged_url_re $site match site tag]} {
                 lappend urlmap($tag) $site
             } else {
                 lappend urlmap($listname) $site
@@ -216,10 +222,11 @@ proc portfetch::sortsites {urls default_listvar} {
     global $default_listvar
     upvar $urls fetch_urls
     variable urlmap
+    set hostregex {[a-zA-Z]+://([a-zA-Z0-9\.\-_]+)}
 
     foreach {url_var distfile} $fetch_urls {
         if {![info exists urlmap($url_var)]} {
-            if {$url_var != $default_listvar} {
+            if {$url_var ne $default_listvar} {
                 ui_error [format [msgcat::mc "No defined site for tag: %s, using $default_listvar"] $url_var]
                 set urlmap($url_var) $urlmap($default_listvar)
             } else {
@@ -227,8 +234,6 @@ proc portfetch::sortsites {urls default_listvar} {
             }
         }
         set urllist $urlmap($url_var)
-        set hosts {}
-        set hostregex {[a-zA-Z]+://([a-zA-Z0-9\.\-_]+)}
 
         if {[llength $urllist] <= 1} {
             # there is only one mirror, no need to ping or sort
@@ -242,6 +247,7 @@ proc portfetch::sortsites {urls default_listvar} {
             seteuid 0; setegid 0
         }
 
+        set hosts {}
         foreach site $urllist {
             if {[string range $site 0 6] eq "file://"} {
                 set pingtimes(localhost) 0
@@ -253,24 +259,49 @@ proc portfetch::sortsites {urls default_listvar} {
             if { [info exists seen($host)] } {
                 continue
             }
-            if { ![info exists seen($host)] } {
-                # first check the persistent cache
-                set pingtimes($host) [get_pingtime $host]
-                if {$pingtimes($host) == {}} {
-                    if {[catch {set fds($host) [open "|ping -noq -c3 -t3 $host | grep round-trip | cut -d / -f 5"]}]} {
-                        ui_debug "Spawning ping for $host failed"
-                        # will end up after all hosts that were pinged OK but before those that didn't respond
-                        set pingtimes($host) 5000
-                    } else {
-                        set seen($host) yes
-                        lappend hosts $host
-                    }
-                }
+            # first check the persistent cache
+            set pingtimes($host) [get_pingtime $host]
+            if {$pingtimes($host) eq {}} {
+                set seen($host) yes
+                lappend hosts $host
             }
         }
 
+        set max_hosts_to_ping 50
+        set len [llength $hosts]
+        if {$len > $max_hosts_to_ping} {
+            # randomize them
+            # shuffle10a from https://wiki.tcl-lang.org/page/Shuffle+a+list
+            while {$len} {
+                set n [expr {int($len*rand())}]
+                set tmp [lindex $hosts $n]
+                lset hosts $n [lindex $hosts [incr len -1]]
+                lset hosts $len $tmp
+            }
+        }
+
+        set pinged_hosts [list]
         foreach host $hosts {
-            gets $fds($host) pingtimes($host)
+            if {[llength $pinged_hosts] < $max_hosts_to_ping} {
+                if {[catch {set fds($host) [open "|ping -noq -c3 -t3 $host"]}]} {
+                    ui_debug "Spawning ping for $host failed"
+                } else {
+                    lappend pinged_hosts $host
+                    continue
+                }
+            }
+            # will end up after all hosts that were pinged OK but before those that didn't respond
+            set pingtimes($host) 5000
+        }
+
+        foreach host $pinged_hosts {
+            set pingtimes($host) ""
+            while {[gets $fds($host) pingline] >= 0} {
+                if {[string match round-trip* $pingline]} {
+                    set pingtimes($host) [lindex [split $pingline /] 4]
+                    break
+                }
+            }
             if { [catch { close $fds($host) }] || ![string is double -strict $pingtimes($host)] } {
                 # ping failed, so put it last in the list
                 set pingtimes($host) 10000
@@ -284,7 +315,7 @@ proc portfetch::sortsites {urls default_listvar} {
             seteuid $oldeuid
         }
 
-        set pinglist {}
+        set pinglist [list]
         foreach site $urllist {
             if {[string range $site 0 6] eq "file://"} {
                 set host localhost
@@ -309,7 +340,7 @@ proc portfetch::sortsites {urls default_listvar} {
 proc portfetch::get_urls {} {
     variable fetch_urls
     variable urlmap
-    set urls {}
+    set urls [list]
 
     portfetch::checkfiles fetch_urls
 
