@@ -1,6 +1,6 @@
 # -*- coding: utf-8; mode: tcl; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- vim:fenc=utf-8:filetype=tcl:et:sw=4:ts=4:sts=4
 #
-# Copyright (c) 2007 - 2015 The MacPorts Project
+# Copyright (c) 2007 - 2020 The MacPorts Project
 # Copyright (c) 2007 Markus W. Weissmann <mww@macports.org>
 # Copyright (c) 2002 - 2003 Apple Inc.
 # All rights reserved.
@@ -32,11 +32,14 @@
 
 package provide portconfigure 1.0
 package require portutil 1.0
+package require portprogress 1.0
+package require struct::set
 
 set org.macports.configure [target_new org.macports.configure portconfigure::configure_main]
 target_provides ${org.macports.configure} configure
 target_requires ${org.macports.configure} main fetch checksum extract patch
 target_prerun ${org.macports.configure} portconfigure::configure_start
+target_postrun ${org.macports.configure} portconfigure::configure_finish
 
 namespace eval portconfigure {
 }
@@ -234,8 +237,24 @@ options configure.optflags \
 default configure.optflags      -Os
 default configure.cflags        {${configure.optflags}}
 default configure.objcflags     {${configure.optflags}}
-default configure.cppflags      {-I${prefix}/include}
-default configure.ldflags       {-L${prefix}/lib -Wl,-headerpad_max_install_names}
+default configure.cppflags      {[portconfigure::configure_get_cppflags]}
+proc portconfigure::configure_get_cppflags {} {
+    global prefix
+    if {[option compiler.limit_flags]} {
+        return ""
+    } else {
+        return -I${prefix}/include
+    }
+}
+default configure.ldflags       {[portconfigure::configure_get_ldflags]}
+proc portconfigure::configure_get_ldflags {} {
+    global prefix
+    if {[option compiler.limit_flags]} {
+        return -Wl,-headerpad_max_install_names
+    } else {
+        return "-L${prefix}/lib -Wl,-headerpad_max_install_names"
+    }
+}
 default configure.libs          {}
 default configure.fflags        {${configure.optflags}}
 default configure.f90flags      {${configure.optflags}}
@@ -257,11 +276,12 @@ default configure.pkg_config_path   {}
 
 options configure.build_arch configure.ld_archflags \
         configure.sdk_version configure.sdkroot \
-        configure.developer_dir
+        configure.sysroot configure.developer_dir
 default configure.build_arch    {[portconfigure::choose_supported_archs ${build_arch}]}
 default configure.ld_archflags  {[portconfigure::configure_get_ld_archflags]}
 default configure.sdk_version   {$macosx_sdk_version}
 default configure.sdkroot       {[portconfigure::configure_get_sdkroot ${configure.sdk_version}]}
+default configure.sysroot       {[expr {${configure.sdkroot} ne "" ? ${configure.sdkroot} : "/"}]}
 default configure.developer_dir {[portconfigure::configure_get_developer_dir]}
 foreach tool {cc objc f77 f90 fc} {
     options configure.${tool}_archflags
@@ -391,7 +411,7 @@ proc portconfigure::choose_supported_archs {archs} {
 
     if {${configure.sdk_version} ne ""} {
         # Figure out which archs are supported by the SDK
-        if {[vercmp ${configure.sdk_version} 11.0] >= 0} {
+        if {[vercmp ${configure.sdk_version} 11] >= 0} {
             set sdk_archs [list arm64 x86_64]
         } elseif {[vercmp ${configure.sdk_version} 10.14] >= 0} {
             set sdk_archs x86_64
@@ -430,17 +450,21 @@ proc portconfigure::choose_supported_archs {archs} {
         # No SDK version (maybe not on macOS)
         set intersection_archs $supported_archs
     }
-    set ret {}
-    # Filter out unsupported archs, but allow demoting to 32-bit if needed.
-    # That means if build_arch is x86_64 it's still possible to build a port
+    set ret [list]
+    # Filter out unsupported archs, but allow demoting to another arch
+    # supported by the SDK if needed, e.g. 64-bit to 32-bit. That means
+    # e.g. if build_arch is x86_64 it's still possible to build a port
     # that sets supported_archs to "i386 ppc" if the SDK allows it.
+    array set arch_demotions [list \
+                                arm64 x86_64 \
+                                x86_64 i386 \
+                                ppc64 ppc \
+                                i386 ppc]
     foreach arch $archs {
         if {$arch in $intersection_archs} {
             set add_arch $arch
-        } elseif {$arch eq "x86_64" && "i386" in $intersection_archs} {
-            set add_arch "i386"
-        } elseif {$arch eq "ppc64" && "ppc" in $intersection_archs} {
-            set add_arch "ppc"
+        } elseif {[info exists arch_demotions($arch)] && $arch_demotions($arch) in $intersection_archs} {
+            set add_arch $arch_demotions($arch)
         } else {
             continue
         }
@@ -464,7 +488,7 @@ proc portconfigure::configure_get_archflags {tool} {
             $tool in {cc cxx objc objcxx}
         } then {
             set flags "-arch ${configure.build_arch}"
-        } elseif {${configure.build_arch} eq "x86_64" || ${configure.build_arch} eq "ppc64"} {
+        } elseif {${configure.build_arch} in [list arm64 ppc64 x86_64]} {
             set flags "-m64"
         } elseif {${configure.compiler} ne "gcc-3.3"} {
             set flags "-m32"
@@ -486,8 +510,22 @@ proc portconfigure::configure_get_ld_archflags {} {
     }
 }
 
+# find a "close enough" match for the given sdk_version in sdk_path
+proc portconfigure::find_close_sdk {sdk_version sdk_path} {
+    # only works right for versions >= 11, which is all we need
+    set sdk_major [lindex [split $sdk_version .] 0]
+    set sdks [glob -nocomplain -directory $sdk_path MacOSX${sdk_major}*.sdk]
+    foreach sdk [lsort -command vercmp $sdks] {
+        # Sanity check - mostly empty SDK directories are known to exist
+        if {[file exists ${sdk}/usr/include/sys/cdefs.h]} {
+            return $sdk
+        }
+    }
+    return ""
+}
+
 proc portconfigure::configure_get_sdkroot {sdk_version} {
-    global developer_dir macosx_version xcodeversion os.arch os.major os.platform use_xcode
+    global developer_dir macos_version_major xcodeversion os.arch os.major os.platform use_xcode
 
     # This is only relevant for macOS
     if {${os.platform} ne "darwin"} {
@@ -495,33 +533,47 @@ proc portconfigure::configure_get_sdkroot {sdk_version} {
     }
 
     # Special hack for Tiger/ppc, since the system libraries do not contain intel slices
-    if {${os.arch} eq "powerpc" && $macosx_version eq "10.4" && [variant_exists universal] && [variant_isset universal]} {
+    if {${os.arch} eq "powerpc" && $macos_version_major eq "10.4" && [variant_exists universal] && [variant_isset universal]} {
         return ${developer_dir}/SDKs/MacOSX10.4u.sdk
     }
 
     # Use the DevSDK (eg: /usr/include) if present and the requested SDK version matches the host version
-    if {${os.major} < 19 && $sdk_version eq $macosx_version && [file exists /usr/include/sys/cdefs.h]} {
+    if {${os.major} < 19 && $sdk_version eq $macos_version_major && [file exists /usr/include/sys/cdefs.h]} {
         return {}
     }
 
+    set sdk_major [lindex [split $sdk_version .] 0]
     set cltpath /Library/Developer/CommandLineTools
     # Check CLT first if Xcode shouldn't be used
     if {![tbool use_xcode]} {
         set sdk ${cltpath}/SDKs/MacOSX${sdk_version}.sdk
         if {[file exists $sdk]} {
             return $sdk
+        } elseif {$sdk_major >= 11} {
+            # SDKs have minor versions as of macOS 11
+            set sdk [find_close_sdk $sdk_version ${cltpath}/SDKs]
+            if {$sdk ne ""} {
+                return $sdk
+            }
         }
 
-        if {[info exists ::portconfigure::sdkroot_cache(macosx${sdk_version})]} {
-            if {$::portconfigure::sdkroot_cache(macosx${sdk_version}) ne ""} {
-                return $::portconfigure::sdkroot_cache(macosx${sdk_version})
-            }
-            # negative result cached, do nothing here
-        } elseif {![catch {exec env DEVELOPER_DIR=${cltpath} xcrun --sdk macosx${sdk_version} --show-sdk-path 2> /dev/null} sdk]} {
-            set ::portconfigure::sdkroot_cache(macosx${sdk_version}) $sdk
-            return $sdk
+        if {$sdk_major >= 11 && $sdk_major == $macos_version_major} {
+            set try_versions [list ${sdk_major}.0 [option macos_version]]
         } else {
-            set ::portconfigure::sdkroot_cache(macosx${sdk_version}) ""
+            set try_versions [list $sdk_version]
+        }
+        foreach try_version $try_versions {
+            if {[info exists ::portconfigure::sdkroot_cache(macosx${try_version})]} {
+                if {$::portconfigure::sdkroot_cache(macosx${try_version}) ne ""} {
+                    return $::portconfigure::sdkroot_cache(macosx${try_version})
+                }
+                # negative result cached, do nothing here
+            } elseif {![catch {exec env DEVELOPER_DIR=${cltpath} xcrun --sdk macosx${try_version} --show-sdk-path 2> /dev/null} sdk]} {
+                set ::portconfigure::sdkroot_cache(macosx${try_version}) $sdk
+                return $sdk
+            } else {
+                set ::portconfigure::sdkroot_cache(macosx${try_version}) ""
+            }
         }
 
         # Fallback on "macosx"
@@ -557,23 +609,42 @@ proc portconfigure::configure_get_sdkroot {sdk_version} {
 
     if {[file exists $sdk]} {
         return $sdk
+    } elseif {$sdk_major >= 11} {
+        # SDKs have minor versions as of macOS 11
+        set sdk [find_close_sdk $sdk_version ${sdks_dir}]
+        if {$sdk ne ""} {
+            return $sdk
+        }
     }
 
-    if {[info exists ::portconfigure::sdkroot_cache(macosx${sdk_version},noclt)]} {
-        if {$::portconfigure::sdkroot_cache(macosx${sdk_version},noclt) ne ""} {
-            return $::portconfigure::sdkroot_cache(macosx${sdk_version},noclt)
-        }
-        # negative result cached, do nothing here
-    } elseif {![catch {exec xcrun --sdk macosx${sdk_version} --show-sdk-path 2> /dev/null} sdk]} {
-        set ::portconfigure::sdkroot_cache(macosx${sdk_version},noclt) $sdk
-        return $sdk
+    if {$sdk_major >= 11 && $sdk_major == $macos_version_major} {
+        set try_versions [list ${sdk_major}.0 [option macos_version]]
     } else {
-        set ::portconfigure::sdkroot_cache(macosx${sdk_version},noclt) ""
+        set try_versions [list $sdk_version]
+    }
+    foreach try_version $try_versions {
+        if {[info exists ::portconfigure::sdkroot_cache(macosx${try_version},noclt)]} {
+            if {$::portconfigure::sdkroot_cache(macosx${try_version},noclt) ne ""} {
+                return $::portconfigure::sdkroot_cache(macosx${try_version},noclt)
+            }
+            # negative result cached, do nothing here
+        } elseif {![catch {exec xcrun --sdk macosx${try_version} --show-sdk-path 2> /dev/null} sdk]} {
+            set ::portconfigure::sdkroot_cache(macosx${try_version},noclt) $sdk
+            return $sdk
+        } else {
+            set ::portconfigure::sdkroot_cache(macosx${try_version},noclt) ""
+        }
     }
 
     set sdk ${cltpath}/SDKs/MacOSX${sdk_version}.sdk
     if {[file exists $sdk]} {
         return $sdk
+    } elseif {$sdk_major >= 11} {
+        # SDKs have minor versions as of macOS 11
+        set sdk [find_close_sdk $sdk_version ${cltpath}/SDKs]
+        if {$sdk ne ""} {
+            return $sdk
+        }
     }
 
     set sdk ${sdks_dir}/MacOSX.sdk
@@ -597,8 +668,8 @@ proc portconfigure::configure_get_sdkroot {sdk_version} {
         return $sdk
     }
 
-    ui_error "Unable to determine location of a macOS SDK."
-    return -code error "Unable to determine location of a macOS SDK."
+    # We can get here if $sdk_version != $macos_version_major on old OS versions
+    return {}
 }
 
 # internal function to determine DEVELOPER_DIR according to Xcode dependency
@@ -761,9 +832,10 @@ proc portconfigure::max_version {verA verB} {
 #--------------------------------------------------------------------
 #|  C Standard  |   Clang   |  Xcode Clang  |   Xcode   |    GCC    |
 #|------------------------------------------------------------------|
-#| 1989 (C89)   |     -     |        -      |     -     |     -     |
-#| 1999 (C99)   |     -     |        -      |     -     |    4.0    |
-#| 2011 (C11)   |    3.3    |   500.2.75    |    5.0    |    4.9    |
+#| 1989 (C89)   |     -     |       -       |     -     |     -     |
+#| 1999 (C99)   |     -     |       -       |     -     |    4.0    |
+#| 2011 (C11)   |    3.3    |  500.2.75     |    5.0    |    4.9    |
+#| 2017 (C17)   |    6.0    |  1000.11.45.2 |   10.0    |    8.0    |
 #--------------------------------------------------------------------
 #
 # https://clang.llvm.org/cxx_status.html
@@ -773,11 +845,11 @@ proc portconfigure::max_version {verA verB} {
 # https://trac.macports.org/wiki/XcodeVersionInfo
 #--------------------------------------------------------------------
 #| C++ Standard |   Clang   |  Xcode Clang  |   Xcode   |    GCC    |
-#|--------------------------------=---------------------------------|
+#|------------------------------------------------------------------|
 #| 1998 (C++98) |     -     |       -       |     -     |     -     |
 #| 2011 (C++11) |    3.3    |   500.2.75    |    5.0    |   4.8.1   |
-#| 2014 (C++14) |    3.4    |   602         |    6.3    |     5     |
-#| 2017 (C++17) |    5.0    |   902.0.39.1  |    9.3    |     7     |
+#| 2014 (C++14) |    3.4    |   602.0.49    |    6.3    |     5     |
+#| 2017 (C++17) |    5.0    |  1000.11.45.2 |   10.0    |     7     |
 #--------------------------------------------------------------------
 #
 # https://openmp.llvm.org
@@ -793,6 +865,21 @@ proc portconfigure::max_version {verA verB} {
 #|      4.5       | Partial |    Future?    | Future? |   ???   |
 #----------------------------------------------------------------
 #
+# https://trac.macports.org/wiki/CompilerEnvironmentVariables
+# https://trac.macports.org/wiki/CompilerSelection#EnvironmentVariables
+#--------------------------------------------------------------------
+#|   Environment Variable   | Xcode Clang | Xcode GCC | Clang | GCC |
+#|------------------------------------------------------------------|
+#| CPATH                    |   318.0.4   |    all    |  all  | all |
+#| LIBRARY_PATH             |  421.0.57   |    all    |  all  | all |
+#| MACOSX_DEPLOYMENT_TARGET |     all     |    all    |  all  | all |
+#| SDKROOT                  | OS X 10.9*  |   none    |  all  |  7  |
+#| DEVELOPER_DIR            | OS X 10.9*  |    N/A    |  N/A  | N/A |
+#| CC_PRINT_OPTIONS         |     all     |    all    |  all  | all |
+#| CC_PRINT_OPTIONS_FILE    |     all     |    all    |  all  | all |
+#--------------------------------------------------------------------
+# * /usr/lib/libxcselect.dylib exists
+#
 # utility procedure: get minimum command line compilers version based on restrictions
 proc portconfigure::get_min_command_line {compiler} {
     global compiler.c_standard compiler.cxx_standard compiler.openmp_version compiler.thread_local_storage os.major
@@ -816,23 +903,35 @@ proc portconfigure::get_min_command_line {compiler} {
                 # no Xcode clang can build against libc++ on < 10.7
                 return none
             }
-            if {${compiler.c_standard} >= 2011} {
+            if {${compiler.c_standard} >= 2017} {
+                set min_value [max_version $min_value 1000.11.45.2]
+            } elseif {${compiler.c_standard} >= 2011} {
                 set min_value [max_version $min_value 500.2.75]
             }
             if {${compiler.cxx_standard} >= 2017} {
-                set min_value [max_version $min_value 902.0.39.1]
+                set min_value [max_version $min_value 1000.11.45.2]
             } elseif {${compiler.cxx_standard} >= 2014} {
-                set min_value [max_version $min_value 602]
+                set min_value [max_version $min_value 602.0.49]
             } elseif {${compiler.cxx_standard} >= 2011} {
-                if {${compiler.thread_local_storage}} {
-                    # macOS has supported thread-local storage since Mac OS X Lion.
-                    # So __thread (GNU extension) and _Thread_local (C11) could be used.
-                    # However, the C++11 keyword was not supported until Xcode 8
-                    #    (https://developer.apple.com/videos/play/wwdc2016-405/?time=354).
-                    set min_value [max_version $min_value 800.0.38]
-                } else {
-                    set min_value [max_version $min_value 500.2.75]
-                }
+                set min_value [max_version $min_value 500.2.75]
+            }
+            if {${compiler.cxx_standard} >= 2011 && ${compiler.thread_local_storage}} {
+                # macOS has supported thread-local storage since Mac OS X Lion.
+                # So __thread (GNU extension) and _Thread_local (C11) could be used.
+                # However, the C++11 keyword was not supported until Xcode 8
+                #    (https://developer.apple.com/library/archive/releasenotes/DeveloperTools/RN-Xcode/Chapters/Introduction.html#//apple_ref/doc/uid/TP40001051-CH1-SW127)
+                #    (https://developer.apple.com/videos/play/wwdc2016-405/?time=354).
+                set min_value [max_version $min_value 800.0.38]
+            }
+            if {[option compiler.limit_flags] || [option compiler.support_environment_paths]} {
+                set min_value [max_version $min_value 421.0.57]
+            }
+            if {
+                ([option compiler.limit_flags] || [option compiler.support_environment_sdkroot]) &&
+                [option configure.sdkroot] ne "" &&
+                !([file exists /usr/lib/libxcselect.dylib] || ${os.major} >= 20)
+            } {
+                return none
             }
         }
         llvm-gcc-4.2 -
@@ -842,6 +941,9 @@ proc portconfigure::get_min_command_line {compiler} {
             if {${compiler.c_standard} > 1999 || ${compiler.cxx_standard} >= 2011 || [option configure.cxx_stdlib] eq "libc++" || ${compiler.thread_local_storage}} {
                 return none
             }
+            if {([option compiler.limit_flags] || [option compiler.support_environment_sdkroot]) && [option configure.sdkroot] ne ""} {
+                return none
+            }
         }
         default {
             return -code error "don't recognize compiler \"${compiler}\""
@@ -854,7 +956,9 @@ proc portconfigure::get_min_command_line {compiler} {
 proc portconfigure::get_min_clang {} {
     global compiler.c_standard compiler.cxx_standard compiler.openmp_version compiler.thread_local_storage
     set min_value 1.0
-    if {${compiler.c_standard} >= 2011} {
+    if {${compiler.c_standard} >= 2017} {
+        set min_value [max_version $min_value 6.0]
+    } elseif {${compiler.c_standard} >= 2011} {
         set min_value [max_version $min_value 3.1]
     }
     if {${compiler.cxx_standard} >= 2017} {
@@ -898,7 +1002,9 @@ proc portconfigure::get_min_gcc {} {
     }
 
     set min_value 1.0
-    if {${compiler.c_standard} >= 2011} {
+    if {${compiler.c_standard} >= 2017} {
+        set min_value [max_version $min_value 8.0]
+    } elseif {${compiler.c_standard} >= 2011} {
         set min_value [max_version $min_value 4.3]
     }  elseif {${compiler.c_standard} >= 1999} {
         set min_value [max_version $min_value 4.0]
@@ -925,7 +1031,12 @@ proc portconfigure::get_min_gcc {} {
         # GCC emulates thread-local storage, but it seems to be broken on older versions of GCC
         set min_value [max_version $min_value 4.5]
     }
-
+    if {
+        ([option compiler.limit_flags] || [option compiler.support_environment_sdkroot]) &&
+        [option configure.sdkroot] ne ""
+    } {
+        set min_value [max_version $min_value 7.0]
+    }
     return ${min_value}
 }
 # utility procedure: get minimum Gfortran version based on restrictions
@@ -945,6 +1056,12 @@ proc portconfigure::get_min_gfortran {} {
         # GCC emulates thread-local storage, but it seems to be broken on older versions of GCC
         set min_value [max_version $min_value 4.5]
     }
+    if {
+        ([option compiler.limit_flags] || [option compiler.support_environment_sdkroot]) &&
+        [option configure.sdkroot] ne ""
+    } {
+        set min_value [max_version $min_value 7.0]
+    }
     return ${min_value}
 }
 #
@@ -958,6 +1075,12 @@ proc portconfigure::g95_ok {} {
         # G95 does not support OpenMP
         return no
     }
+    if {
+        ([option compiler.limit_flags] || [option compiler.support_environment_sdkroot]) &&
+        [option configure.sdkroot] ne ""
+    } {
+        return no
+    }
     return yes
 }
 # utility procedure: get Apple compilers based on Xcode version
@@ -1529,32 +1652,34 @@ proc portconfigure::configure_main {args} {
         global configure.${flags} configure.universal_${flags}
     }
 
+    set callback [list "-callback" portprogress::target_progress_callback]
+
     if {[tbool use_autoreconf]} {
-        if {[catch {command_exec autoreconf} result]} {
+        if {[catch {command_exec {*}${callback} autoreconf} result]} {
             return -code error "[format [msgcat::mc "%s failure: %s"] autoreconf $result]"
         }
     }
 
     if {[tbool use_automake]} {
-        if {[catch {command_exec automake} result]} {
+        if {[catch {command_exec {*}${callback} automake} result]} {
             return -code error "[format [msgcat::mc "%s failure: %s"] automake $result]"
         }
     }
 
     if {[tbool use_autoconf]} {
-        if {[catch {command_exec autoconf} result]} {
+        if {[catch {command_exec {*}${callback} autoconf} result]} {
             return -code error "[format [msgcat::mc "%s failure: %s"] autoconf $result]"
         }
     }
 
     if {[tbool use_xmkmf]} {
         parse_environment xmkmf
-        if {[catch {command_exec xmkmf} result]} {
+        if {[catch {command_exec {*}${callback} xmkmf} result]} {
             return -code error "[format [msgcat::mc "%s failure: %s"] xmkmf $result]"
         }
 
         parse_environment xmkmf
-        if {[catch {command_exec "cd ${worksrcpath} && make Makefiles" -varprefix xmkmf} result]} {
+        if {[catch {command_exec {*}${callback} "cd ${worksrcpath} && make Makefiles" -varprefix xmkmf} result]} {
             return -code error "[format [msgcat::mc "%s failure: %s"] "make Makefiles" $result]"
         }
     } elseif {[tbool use_configure]} {
@@ -1611,7 +1736,7 @@ proc portconfigure::configure_main {args} {
         }
 
         # add SDK flags if cross-compiling (or universal on ppc tiger)
-        if {${configure.sdkroot} ne ""} {
+        if {${configure.sdkroot} ne "" && ![option compiler.limit_flags]} {
             foreach env_var {CPPFLAGS CFLAGS CXXFLAGS OBJCFLAGS OBJCXXFLAGS} {
                 append_to_environment_value configure $env_var -isysroot${configure.sdkroot}
             }
@@ -1644,13 +1769,92 @@ proc portconfigure::configure_main {args} {
         }
 
         # Execute the command (with the new environment).
-        if {[catch {command_exec configure} result]} {
-            global configure.dir
-            if {[file exists ${configure.dir}/config.log]} {
-                ui_error "[format [msgcat::mc "Failed to configure %s, consult %s/config.log"] [option subport] ${configure.dir}]"
+        if {[catch {command_exec {*}${callback} configure} result]} {
+            global configure.dir build.dir
+            foreach error_log [list ${configure.dir}/config.log ${configure.dir}/CMakeFiles/CMakeError.log ${build.dir}/meson-logs/meson-log.txt] {
+                if {[file exists ${error_log}]} {
+                    ui_error "[format [msgcat::mc "Failed to configure %s: consult %s"] [option subport] ${error_log}]"
+                }
             }
             return -code error "[format [msgcat::mc "%s failure: %s"] configure $result]"
         }
     }
     return 0
 }
+
+options configure.checks.implicit_function_declaration \
+        configure.checks.implicit_function_declaration.whitelist
+default configure.checks.implicit_function_declaration yes
+default configure.checks.implicit_function_declaration.whitelist {[portconfigure::load_implicit_function_declaration_whitelist ${configure.sdk_version}]}
+
+proc portconfigure::check_implicit_function_declarations {} {
+    global \
+        workpath \
+        configure.checks.implicit_function_declaration.whitelist
+
+    # Map from function name to config.log that used it without declaration
+    array set undeclared_functions {}
+
+    fs-traverse -tails file [list ${workpath}] {
+        if {[file tail $file] in [list config.log CMakeError.log meson-log.txt] && [file isfile [file join ${workpath} $file]]} {
+            # We could do the searching ourselves, but using a tool optimized for this purpose is likely much faster
+            # than using Tcl.
+            #
+            # Using /usr/bin/fgrep here, so we don't accidentally pick up a macports-installed grep which might
+            # currently not be runnable due to a missing library.
+            set args [list "/usr/bin/fgrep" "--" "-Wimplicit-function-declaration"]
+            lappend args [file join ${workpath} $file]
+
+            if {![catch {set result [exec -- {*}$args]}]} {
+                foreach line [split $result "\n"] {
+                    if {[regexp -- "(?:implicit declaration of function|implicitly declaring library function) '(\[^']+)'" $line -> function]} {
+                        set is_whitelisted no
+                        foreach whitelisted ${configure.checks.implicit_function_declaration.whitelist} {
+                            if {[string match -nocase $whitelisted $function]} {
+                                set is_whitelisted yes
+                                break
+                            }
+                        }
+                        if {!$is_whitelisted} {
+                            ::struct::set include undeclared_functions($function) $file
+                        } else {
+                            ui_debug [format "Ignoring implicit declaration of function '%s' because it is whitelisted" $function]
+                        }
+                    }
+                }
+            }
+        }
+    }
+
+    if {[array size undeclared_functions] > 0} {
+        ui_warn "Configuration logfiles contain indications of -Wimplicit-function-declaration; check that features were not accidentally disabled:"
+        foreach {function files} [array get undeclared_functions] {
+            ui_msg [format "  %s: found in %s" $function [join $files ", "]]
+        }
+    }
+}
+
+proc portconfigure::load_implicit_function_declaration_whitelist {sdk_version} {
+    set whitelist {}
+
+    set whitelist_file [getdefaultportresourcepath "port1.0/checks/implicit_function_declaration/macosx${sdk_version}.sdk.list"]
+    if {[file exists $whitelist_file]} {
+        set fd [open $whitelist_file r]
+        while {[gets $fd whitelist_entry] >= 0} {
+            lappend whitelist $whitelist_entry
+        }
+        close $fd
+    }
+
+    return $whitelist
+}
+
+proc portconfigure::configure_finish {args} {
+    global \
+        configure.dir \
+        configure.checks.implicit_function_declaration
+
+    if {[file isdirectory ${configure.dir}] && ${configure.checks.implicit_function_declaration}} {
+        portconfigure::check_implicit_function_declarations
+    }
+}
